
/*
 * LZMA test
 */

#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "../Alloc.h"
#include "../7zFile.h"
#include "../7zVersion.h"
#include "../LzmaDec.h"
#include "../LzmaEnc.h"
#include "../LzmaLib.h"


#define WARC_HEADER "warc/0.9 "
#define warcLen 9
#define rowMax 10000000
#define numTests 10


typedef unsigned long uLong;
typedef unsigned int  uInt;


// In millisecs
double getTime(void)
{

	double usertime, systime;
	struct rusage usage;

	getrusage(RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec * 1000 +
		(double) usage.ru_utime.tv_usec / 1000.0;
	systime = (double) usage.ru_stime.tv_sec * 1000 +
		(double) usage.ru_stime.tv_usec / 1000.0;
	return (usertime + systime);

}

// -------------------------------------
// If line is longer than lim chars, then it is split
int getLine(Byte *s, int lim)
{
  int c,i;
  
  i=0;
  while((--lim>0) && ((c=getchar())!= EOF) && (c != '\n'))
    s[i++]= (Byte) c;
  if (c == '\n')
    s[i++]= (Byte) c;
  s[i] = '\0'; // just for printing purposes
  return i;
}


// -------------------------------------
// Ease the printing of a prefix
void printPrefix(Byte *s, int num)
{
  for( ; (num > 0) && (*s != '\0'); num--,s++)
    printf("%c",*s);
}

// -------------------------------------
// Returns the number of fetched chars
int getBlock(Byte *b, uLong bSize)
{
  int c;
  uLong len;

  len=0;
  c = 5; //dummy assignment

  while((len < bSize) && (c != EOF)){

    c = getchar(); // fetch next char
    if ( c != EOF ) 
      { b[len++] = (Byte) c; }
  }

  return len;
}


// -------------------------------------
int PrintError(char *buffer, const char *message)
{
  strcat(buffer, "\nError: ");
  strcat(buffer, message);
  strcat(buffer, "\n");
  return 1;
}



int main (int argc, char *argv[])
{
  double getTime(void);
  int getline(Byte *, int);

  Byte *inBuffer, *outBuffer, *row;
  uLong outLen;
  double totCompr, totUnCompr;
  uLong blockSize, posBuf;
  uLong numBlocks, lenTest;
  double startTime, endTime, totTime, timeTests;
  int err, allocatedBuf, lineLen, ii, cOption;
  double minS,maxS,minC,maxC,currSpeed,totSpeed,TmpRatio;

  SRes res;
  size_t  outSize, outSize1;
  FILE *fp;
  unsigned char outProps[LZMA_PROPS_SIZE]; 
  size_t outPropsSize = LZMA_PROPS_SIZE;

  
  // Loading the input parameters
  if (argc < 1)	{
    fprintf(stderr, "Reads from STDIN and needs the blockSize in MBs\n\n");
    exit (1);
  }

  // Creates the buffers (of some min size)
  blockSize = (uLong)(atof(argv[1]) * 1000000); //in bytes
  allocatedBuf = (int)((blockSize + 5000000)  * 1.3);


  // Creates buffers
  inBuffer = (Byte *) calloc((uInt)allocatedBuf,1);
  outBuffer = (Byte *) calloc((uInt)allocatedBuf,1);
  row = (Byte *) malloc(rowMax);
  
  if((inBuffer == NULL) || (outBuffer == NULL) || (row == NULL)){
    fprintf(stderr, "Error in allocating the buffers\n\n");
    exit (1);  }


  totTime = 0.0;
  totCompr = 0;
  totUnCompr = 0;
  numBlocks = 1;
  posBuf = 0;
  minS = 1000;
  maxS = -1;
  maxC = -1;
  minC = 1000;
  totSpeed = 0.0;


  while(lineLen = getLine(row,rowMax)){

    if (lineLen > rowMax) {
      fprintf(stderr,"row too long...\n");
      exit(1);
    }


    // Skip the warc header
    if (strncmp((char *)row, WARC_HEADER, warcLen) == 0){
      int i;
      for(i=0; i < 9; i++)
	getLine(row,rowMax);
      lineLen = getLine(row,rowMax);
    }


    //Copy row in the buffer
    memcpy(inBuffer+posBuf, row, lineLen);
    posBuf += lineLen;

    // Check overflow
    if(posBuf > allocatedBuf){
      fprintf(stderr,"Buffer too much loaded...\n");
      exit(1);
    }



    if (posBuf > blockSize){


      // Compress
      outSize = allocatedBuf;
      totUnCompr += (posBuf / 1000000.0);

      // Description of the parameters in ../LzmaLib.h
      // Dict size = 32Mb, compression level = 9, other params = default  
      res = LzmaCompress(outBuffer, &outSize, inBuffer, posBuf, outProps, &outPropsSize, 9, (1<<24), 3, 0, 2, 32, 2);

      if (res != SZ_OK)
	{ fprintf(stderr, "Error in LZMA Compress: %d\n",res); exit(1); }

      totCompr += (outSize / 1000000.0);


      // timings: millisecs
      startTime = getTime();
      for(ii=0; ii < numTests; ii++) {
      
	// Decompression
	// ATTENTION: outSize1 must be the final uncompressed output length
	//            so it must be passed someway.
	outSize1 = (size_t)posBuf;  
	res = LzmaUncompress(inBuffer, &outSize1, outBuffer, &outSize, outProps, outPropsSize);
	if (res != SZ_OK)
	  { fprintf(stderr, "Error in LZMA Decompress: %d\n",res); exit(1); }

	if (posBuf != outSize1){
	  fprintf(stderr, "Error in decompression (len)\n\n");
	  exit(1);  }
      }
      
      // totTime in millisecs --> hence we get MB/sec
      endTime = getTime();      
      timeTests = (endTime - startTime);
      
      // Update counters: in milliseconds
      totTime += (timeTests / (1000.0 * numTests));

      currSpeed = (posBuf * numTests) / (1000.0 * timeTests);
      totSpeed += currSpeed;

      minS = (currSpeed < minS ? currSpeed : minS);
      maxS = (currSpeed > maxS ? currSpeed : maxS);

      minC = ((100.0 * outSize)/posBuf < minC ? (100.0 * outSize)/posBuf : minC);
      maxC = ((100.0 * outSize)/posBuf > maxC ? (100.0 * outSize)/posBuf : maxC);

      TmpRatio = 100.0 * ((1.0 * totCompr)/totUnCompr);

      printf("\n... this block %.3f MB/sec, %.2lf comprRatio\n", currSpeed, (100.0 * outSize)/posBuf);
      printf("%.2f totUncompr MB, %.2lf TOTcomprRatio, minC %.2lf perc., maxC %.2lf perc.\n", totUnCompr, TmpRatio,minC,maxC);
      printf("%d blocks at (an avg of) %.2lf MB/sec, minS %.3lf MB/s, maxS %.3lf MB/s\n", 
	     numBlocks, totSpeed/numBlocks,minS,maxS);


	numBlocks++;
	posBuf = 0; //restart buffer
    }
  }
    
  free(inBuffer);
  free(outBuffer);
  return 0;
}


