/* 

   March 2009, Ferragina Paolo (ferragina@di.unipi.it)
   Copyright @ Paolo Ferragina.

   bmf_compressor.c -- This is a combination of Bentley-McIlroy's
   compressor, with Gzip or Lzma compressors as post-compression
   tools. The former combination implements the idea published in
   BigTable's paper [Procs OSDI '06] by Google's researchers.  The
   latter is my idea which deploys the powerful lzma's compressor by
   I. Pavlov, and still keeps about the fast decompression performance
   of gzip.

   Usage: bmfC <CompressorType: g|l|n> <ChunkSizeToCopy> <PagesPerBlock> <OutputFile>
   
   Reads from STDIN a stream which is formatted according to the
   WARC-format.  It then partitions the file into blocks of #pages =
   <PagesPerBlock>, and on each block it applies the BmI-algorithm
   which searches for repeated <ChunkSizeToCopy> bytes far back in the
   input.  Each BMI-compressed block is then post-compressed by using
   gzip/lzma or none, according to the first input parameter. The compressed
   output is written to two compressed files with extensions
   .bmf.offset and .bmf.compr. The former keeps the offsets in bytes
   of the compressed blocks.  The latter keeps the compressed blocks
   which, therefore, have variable size.
   
   FORMAT: 
   .bmf.offset = comprType (1b) + pagesBlock (4b) + offsets (8b each)....
   .bmf.pages  = compr (byte sequence)

   In the case of Lzma, and for the purpose of decompression, each
   block is prefixed by an integer (4 bytes) indicating the length of
   the uncompressed version of that block.  We mention that
   <PagesPerBlock> must be chosen to make the overall occupied space
   by each block smaller than 300Mb. This is just for simplicity.

   NOTICE: This code can be used only for research purposes. All rights
   are owned by the author: Paolo Ferragina, Dipartimento di Informatica,
   University of Pisa, Italy.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "ZLIB/zlib.h"
#include "libbmi.h"

#include "LZMA/Alloc.h"
#include "LZMA/7zFile.h"
#include "LZMA/7zVersion.h"
#include "LZMA/LzmaDec.h"
#include "LZMA/LzmaEnc.h"
#include "LZMA/LzmaLib.h"

#include "aux_funct.h"



/* ---------------- MAIN TYPES AND DATA ---------------------- */

typedef unsigned long long uLongLong; 


#define MAXB  1000
#define MAXN  300000000
#define WARC_HEADER "warc/0.9 "
#define warcLen 9
#define rowMax 10000000


/* ---------------------  MAIN ----------------------- */


int main(int argc, char *argv[])
{	

  /* ---- First compression step */
  unsigned char *text;			       /* Text to compress: array */
  int textSize;			       /* Text to compress: size */
  int blockSize;		       /* substring to hash: (minimum) size */

  /* ---- Second compression step */
  unsigned char *bmiBuffer, *bmigzBuffer;  /* Array where to store the compressed data */
  int  bmiLen, bmigzLen;        /* Len of previous arrays */
  FILE *outOff, *outCompr; /* Files where to write the compressed data and offsets */


  /* ---- Our management of the output */
  char *outFileName, tmpName[256];
  char *row, ComprType;
  int PagesBlock, posText, numPages;
  uLongLong numBlocks, numTotPages, offsetCompress;
  int lineLen, notLastBlock, err;
  double minC, maxC, totCompr, totUnCompr, tmp=0;
  double startTime, endTime;
  
  /* ----- LZMA variables */
  SRes res;
  unsigned char outProps[LZMA_PROPS_SIZE];
  size_t outPropsSize = LZMA_PROPS_SIZE;


  /* Arguments check */
  if(argc < 4){
    printf("\n\nSyntax:\n bmfC <ComprType: g|l|n> <MinSizeChunkToCopy> <PagesPerBlock> <outputFile>\n\n");
    exit(1);
  } else {
    printf("\n\n");
  }
  

  /* PostCompressor type */
  ComprType = argv[1][0];
  if ((ComprType != 'g') && (ComprType != 'l' ) && (ComprType != 'n')){
    fprintf(stderr,"You need to specify either g (gzip) or l (lzma) or n (none)\n\n");
    exit(1); }
  

  /* Size of the chunk to hash */
  blockSize = atoi(argv[2]);
  if ((blockSize <= 0) || (blockSize > MAXB)){
    fprintf(stderr,"Invalid size for the Chunk to be hashed: 1 - %d\n\n", MAXB);
    exit(1); }
  

  /* Num pages per block to scan and compress
     We assume an average size of 100K and a default of 10Mb */
  PagesBlock = atoi(argv[3]);
  if ((PagesBlock <= 0) || (PagesBlock > (MAXN/100000) )){
    fprintf(stderr,"Invalid number of pages to compress: 1 - %d\n\n",MAXN/100000);
    exit(1); }
  textSize = (int)MAXN;
  text = (unsigned char *) malloc(textSize + 1);
  if (text == NULL){
    fprintf(stderr,"\n\nError in allocating the memory for the block to compress\n");
    exit(1); }

  
  /* Opening the two output files */
  outFileName  = argv[4];
  sprintf(tmpName,"%s.bmf.offset",outFileName);
  outOff = fopen(tmpName,"w+");
  sprintf(tmpName,"%s.bmf.compr",outFileName);
  outCompr = fopen(tmpName,"w+");
  if ((outOff == NULL) || (outCompr == NULL)){
    fprintf(stderr,"\n\nError in opening the output files\n");
    exit(1); }
  

  /* BMI buffer
     Tipically bmi will shrink the file. We prefer this static allocation
     to avoid millions of dynamic memory allocations, one per block */
  bmiLen = (int) (textSize * 1.1);
  bmiBuffer = (unsigned char *) malloc(bmiLen);
  if(bmiBuffer == NULL){
    fprintf(stderr, "Error in allocating the bmiBuffer\n\n");
    exit (1);  }


  /* We could recycle the buffer of the text to save space but we
     prefer to be safe in allocation */
  bmigzLen = (int) (textSize * 1.1);
  bmigzBuffer = (unsigned char *) malloc(bmigzLen);
  if(bmigzBuffer == NULL){
    fprintf(stderr, "Error in allocating the bmigzBuffer\n\n");
    exit (1);  }


  // ---- WARC preamble ----
  // Skip dummy rows before WARC header
  row = (char *) malloc(rowMax);
  if(row == NULL){
    fprintf(stderr, "Error in allocating the row\n\n");
    exit (1);  }
  row[0] = 'x'; lineLen=0; // dummy initialization
  while(strncmp(row,WARC_HEADER,warcLen) != 0)
    lineLen=getLine(row,rowMax);
  if (strncmp(row,WARC_HEADER,warcLen) != 0){
    fprintf(stderr,"Cannot find first page!\n\n");
    exit(1);}


  // Some statistics
  maxC = -1;  minC = 10000;  // max/min compression ratio (dummy values)
  totCompr = 0.0;            // total compression ratio
  totUnCompr = 0;            // total uncompressed size (in  MB)


  numBlocks = 0;        // #blocks compressed
  numPages = 0;         // #pages in current block
  numTotPages = 0;      // #pages in total
  notLastBlock = 1;     // to force compressing last (unfull) block
  posText = 0;          // byte iterator over the current block
  offsetCompress = 0L;  // byte offset within the bmf.compr file


  startTime = getTime();

  // Preamble of the .bmf.offset file
  fwrite(&ComprType,sizeof(char),  1, outOff);
  fwrite(&PagesBlock,sizeof(int),  1, outOff);


  // start fetching blocks of PagesBlock pages each
  while((lineLen > 0) || (notLastBlock)){
    
    if (lineLen > 0){
      if (strncmp(row,WARC_HEADER,warcLen) == 0)
	{ numPages++; numTotPages++; }
    } else { // lineLen == 0 --> we need to manage the last block
      notLastBlock = 0;
    }

    // We manage both cases: full block or last block
    if ((numPages > PagesBlock) || (notLastBlock == 0)){
      
      if (text[posText] != 255) // Last two pos must differ: BmI's specialty
	text[posText] = (char) 255; 
      else
	text[posText] = (char) 254;


      if(posText+1 > textSize){
	fprintf(stderr,"Loaded to much pages in buffer!\n");
	exit(1); }
      

      // BMI-compression
      bmiLen = (int)(textSize * 1.1);
      bmi_compress(bmiBuffer, &bmiLen, text, posText, blockSize);


      // post-compression: either gzip or lzma
      if (ComprType == 'l') { // This is LZMA

	// Dict size = 32Mb, compression level = 9, other params = default
	bmigzLen = (int) (textSize * 1.1);
	res = LzmaCompress(bmigzBuffer, (size_t *) &bmigzLen, bmiBuffer, (size_t) bmiLen, 
			   outProps, &outPropsSize, 9, (1<<24), 3, 0, 2, 32, 2);
	if (res != SZ_OK) { fprintf(stderr, "Error in LZMA Compress: %d\n",res); exit(1);}


	// Write on file
	fwrite(&offsetCompress, sizeof(uLongLong),  1, outOff);
	fwrite(&bmiLen, sizeof(int), 1, outCompr);  // lzmaDEcompr needs to know it
	fwrite(&outPropsSize, sizeof(char), 1, outCompr);
	fwrite(outProps, sizeof(char), outPropsSize, outCompr);
	fwrite(bmigzBuffer, sizeof(char), bmigzLen, outCompr);

	offsetCompress += (uLongLong)(sizeof(int) + outPropsSize 
				      * sizeof(char) + sizeof(char) + bmigzLen);

      } else if (ComprType == 'g')  { // This is GZIP

	bmigzLen = (int) (textSize * 1.1);
	err = compress2((Byte *)bmigzBuffer, (uLongf *) &bmigzLen, (Byte *) bmiBuffer, bmiLen, 9);
	CHECK_ERR(err, "compress");

	// Write on file
	fwrite(&offsetCompress, sizeof(uLongLong),  1, outOff);
	fwrite(bmigzBuffer, sizeof(unsigned char), (size_t) bmigzLen, outCompr);

	offsetCompress += bmigzLen;

      } else { // This is none post-compression (i.e. bmi-only)

	// Write on file
	fwrite(&offsetCompress, sizeof(uLongLong),  1, outOff);
	fwrite(bmiBuffer, sizeof(unsigned char), (size_t) bmiLen, outCompr);

	// for uniformity with next code
	bmigzLen = bmiLen;
	offsetCompress += bmigzLen;

      }
	

      // statistics
      totUnCompr += (posText / 1000000.0);
      tmp = bmigzLen+8;
      totCompr += (tmp / 1000000.0);
      minC = ((100.0 * tmp)/posText < minC ? (100.0 * tmp)/posText : minC);
      maxC = ((100.0 * tmp)/posText > maxC ? (100.0 * tmp)/posText : maxC);
      tmp = (100.0 * totCompr)/totUnCompr;

      if (numBlocks % 20 == 0){
	fprintf(stderr,"Uncompr. size = %.2lf MB, Ratio: tot %.2f perc, minC %.2lf perc, maxC %.2lf perc\n",totUnCompr,tmp,minC,maxC);
      }

      // Updates the parameters of the block
      numBlocks++;
      numPages = 1;
      posText = 0;

    }

    // ---- In all cases ----
    //Copy the buffered row in the text to compress
    memcpy(text+posText, row, lineLen);
    posText += lineLen;
    
    // fetch the next line
    lineLen = getLine(row,rowMax);
  }


  // Writes the dummy offset of the end-file,
  // which simplifies the subsequent decompr phase
  fwrite(&offsetCompress,sizeof(uLongLong),  1, outOff);

  fclose(outCompr);
  fclose(outOff);
  free(text);
  //free(bmiBuffer);

  endTime = getTime();

  fprintf(stderr,"\n\n---------- SUMMARY:\n");
  fprintf(stderr,"Compression took %.2f secs, at %.3f MB/sec\n", (endTime - startTime)/ 1000.0, (totUnCompr / (endTime - startTime)) * 1000.0);
  fprintf(stderr,"Ratio: tot %.2f perc, minC %.2lf perc, maxC %.2lf perc\n\n",tmp,minC,maxC);

  return 0;
}



  
