/* 

   March 2009, Ferragina Paolo (ferragina@di.unipi.it)

   bmf_extract.c -- This is the decompressor of the algorithm
   specified in bmf_compressor.c. We refer to that file for an explanation.

   Usage: bmfE <InputFile> <-1 | block to decompress>

   It reads the two files: InputFile.bmf.offset, InputFile.bmf.compr.
   the decompressed output is written to the file InputFile.bmf.out.

   NOTICE: This code can be used only for research purposes. All rights
   are owned by the author: Paolo Ferragina, Dipartimento di Informatica,
   University of Pisa, Italy.

*/


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "ZLIB/zlib.h"
#include "libbmi.h"

#include "LZMA/Alloc.h"
#include "LZMA/7zFile.h"
#include "LZMA/7zVersion.h"
#include "LZMA/LzmaDec.h"
#include "LZMA/LzmaEnc.h"
#include "LZMA/LzmaLib.h"

#include "aux_funct.h"



// We assume that each block to be compressed is smaller than 100Mb
#define MAXN 300000000
#define WARC_HEADER "warc/0.9 "
#define warcLen 9

typedef unsigned long long uLongLong;




// Download the block number blkNum (counting from 0), storing it 
// in the byte array *blk, where *blkSize is its length in bytes.
// returned value = -1 if error, 0 otherwise
int getCompressedBlock(FILE *fOff, FILE *fCom, uLongLong blkNum, unsigned char *blk, int *blkSize)
{
  uLongLong currOffset, nextOffset;
  uLongLong where = blkNum * sizeof(uLongLong) + sizeof(int) + sizeof(char);
 
  fseeko64(fOff, (off64_t)where, SEEK_SET);
  fread(&currOffset,sizeof(uLongLong),1,fOff);
  fread(&nextOffset,sizeof(uLongLong),1,fOff); // it surely exists

  // Read the offset to the blk
  fseeko64(fCom, currOffset, SEEK_SET);
  if (nextOffset - currOffset > (unsigned int)*blkSize){
    fprintf(stderr,"Blk too long to be loaded (%lld)\n", nextOffset - currOffset);
    return -1;  }

  // Read the phrase&compr data
  *blkSize = (int) (nextOffset - currOffset);
  size_t tmp = fread(blk, sizeof(char), (size_t) *blkSize, fCom);

  if(tmp != (size_t) *blkSize){
    fprintf(stderr,"Error in reading the block (%d vs %d)\n", *blkSize, tmp);
    return -1;
  } else 
    return 0;
}


// Get the page of relative number PageNum, and identifies it via
// its starting and ending byte in the block blk
// returned value is -1 if error, 0 otherwise
int getPage(int PageNum, unsigned char *blk, int blkSize, 
	     int *startPg, int *endPg)
{

  int posBlk = 0;
  int currPg = 0;

  
  *startPg = -1; *endPg = -1;
  while(posBlk < blkSize){

    // Assume beginning of block == beginning of page
    if (strncmp((const char *)(blk + posBlk), WARC_HEADER, warcLen) == 0){
      
      // Start of the searched page
      if (currPg == PageNum)
	*startPg = posBlk;  

      // End of the searched page
      if (currPg == PageNum+1) {
	*endPg = posBlk;
	return 0; }

      // Increment page counter
      currPg++;
    }

    while((posBlk < blkSize) && (blk[posBlk] != '\n'))
      posBlk++;
    posBlk++; // skip the newLine
  }


  // requested the last page
  if (*startPg != -1) {
    *endPg = posBlk;
    return 0; }

  // error
  return -1;
}


// ------------------------  MAIN -------------------------

int main (int argc, char *argv[])
{

  unsigned char *inBuffer, *outBuffer1, *outBuffer2;
  int inLen, outLen1, outLen2;
  uLongLong allocatedBuf, numOffset, lenOffset, totLen;
  long long w;
  double startTime, endTime, totTime;
  int err, PagesBlock;
  char ComprType, tmpName[256];
  FILE *fileOffset, *fileCompress, *fileOutput;
  
  // Vars LZMA
  SRes res;
  unsigned char outProps[LZMA_PROPS_SIZE];
  size_t outPropsSize = LZMA_PROPS_SIZE;



  if(argc<3){
    printf("\n\nSyntax:\n  bmfE <inputfile> [-1 (all bloks) | blockID to decompress]\n\n");
    exit (1);  }


  /* Opening the two input files */
  sprintf(tmpName,"%s.bmf.offset", argv[1]);
  fileOffset = fopen(tmpName,"r");
  sprintf(tmpName,"%s.bmf.compr", argv[1]);
  fileCompress = fopen(tmpName,"r");
  sprintf(tmpName,"%s.bmf.out", argv[1]);
  fileOutput = fopen(tmpName,"w");
  if ((fileOffset == NULL) || (fileCompress == NULL) || (fileOutput == NULL)){
    printf("\n\nError in opening the I/O files\n");
    exit(1); }


  // first char is ComprType, second int is the #pages per block
  fread(&ComprType, sizeof(char), 1, fileOffset);
  fread(&PagesBlock, sizeof(int), 1, fileOffset);
  if ((ComprType != 'g') && (ComprType != 'l') && (ComprType != 'n')){
    fprintf(stderr, "ComprType not recognized!\n");
    exit(1); }

  // Num tot blocks
  fseek(fileOffset, 0L, SEEK_END);
  lenOffset = ftell(fileOffset) - sizeof(int) - sizeof(char); 
  numOffset = (lenOffset / sizeof(uLongLong)) - 1; // last is the file len



  // What to extract
  w = atoll(argv[2]);
  if((w != -1) && ((uLongLong)w > numOffset)){
    printf("We have %lld blocks only!\n",numOffset);
    exit(1);  }


  // Creates the needed buffers
  allocatedBuf = MAXN;
  inLen = allocatedBuf;
  inBuffer = (unsigned char *) malloc(allocatedBuf);
  if(inBuffer == NULL){
    fprintf(stderr, "Error in allocating the IN buffer\n\n");
    exit (1);  }

  outLen1 = allocatedBuf;
  outBuffer1 = (unsigned char *) malloc(allocatedBuf);
  if(outBuffer1 == NULL){
    fprintf(stderr, "Error in allocating the OUT1 buffer\n\n");
    exit (1);  }

  outLen2 = allocatedBuf;
  outBuffer2 = (unsigned char *) malloc(allocatedBuf);
  if(outBuffer2 == NULL){
    fprintf(stderr, "Error in allocating the OUT2 buffer\n\n");
    exit (1);  }

  // Speed Test Block Decompression
  startTime = getTime();
  totLen = 0;


  for(uLongLong ww=0; ww < numOffset; ww++){

    inLen = outLen1 = outLen2 = allocatedBuf;

    // Force block to be retrieved
    if(w != -1) ww = w;

    getCompressedBlock(fileOffset, fileCompress, ww, inBuffer, &inLen);

    if (ComprType == 'g') { // Gunzip
      
      err = uncompress((Byte *)outBuffer1, (uLongf *) &outLen1, (Byte *)inBuffer, inLen);
      CHECK_ERR(err, "uncompress");

      } else if (ComprType == 'l') { // Lzma


        // ATTENTION: outLen1 must be the final uncompressed output length
        //            so it must be passed someway.
	outLen1 = *((int*)inBuffer);
	
	// Reading the preamble of LZMA
	// Jump the Uncompr Len (int) 
	// and read #Props (1) + Props (5)
	outPropsSize = inBuffer[sizeof(int)];
	for(size_t ii=0; ii < outPropsSize; ii++)
	  outProps[ii] = inBuffer[sizeof(int)+1+ii];

	// jump the preamble: Props + LenBlkUncompr
        res = LzmaUncompress(outBuffer1, (size_t *) &outLen1, inBuffer + outPropsSize + 1 + sizeof(int), (size_t *) &inLen, outProps, outPropsSize);
        if (res != SZ_OK)
          { fprintf(stderr, "Error in LZMA Decompress: %d\n",res); exit(1); }

    } else { // none

      memcpy(outBuffer1, inBuffer, inLen);
      outLen1 = inLen;

    }

      bmi_decompress(outBuffer2, &outLen2, outBuffer1, outLen1);
      totLen += outLen2;

      fwrite(outBuffer2, sizeof(char), outLen2, fileOutput);
      

      

      // Force the exit;
      if (w != -1) ww = numOffset;

      // Statistics
      endTime = getTime();
      totTime = (endTime - startTime) / 1000.0; // avg speed in secs
      if (ww % 20)
	printf("\nnum block: %lld, Tot Uncompr. %.3f MB, decompr. speed: %.3f MB/sec", ww, totLen/1000000.0, (totLen/1000000.0)/totTime);

  }
  

  //timings: millisecs --> hence we get MB/sec
  endTime = getTime();
  totTime = (endTime - startTime) / 1000.0; // avg speed in secs
  printf("\n\n--- Summary:\n");
  printf("Uncompr.  %.3f MB, decompr. speed: %.3f MB/sec\n", totLen/1000000.0, (totLen/1000000.0)/totTime);


  free(inBuffer);
  free(outBuffer1);
  free(outBuffer2);
  fclose(fileOutput);
  fclose(fileOffset);
  fclose(fileCompress);
  
  return 0;
}
