/* 

   March 2009, Ferragina Paolo (ferragina@di.unipi.it)
   bmic.c -- My adaptation of the Bentley-McIlroy's compressor

   Code originally downloaded from: www.cs.dartmouth.edu/~doug/source.html

   My adaptation consists of porting to the GZIP-like interface,
   remove some specialties that made the code a little bit cumbersome
   (for me, at least!), change the copy-encoding to a pair of a
   variable number of bytes which depends on the size of dist&len and
   whose configuration is encoded in the byte following the special
   symbol <. We assume that <0 encodes the symbol <.

   NOTICE: This code can be used only for research purposes. All rights
   are owned by the author: Paolo Ferragina, Dipartimento di Informatica,
   University of Pisa, Italy.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>



// ------------------ AUXILIARY FUNCTIONS --------------------


// Returns the number of bytes needed to encode n
int GetByteLen(int n)
{	
  int tmp = n;
  int len;


  for(len=1; len < 4; len++){
    tmp >>= 8;
    if ((tmp & 0xFF) == 0)
      return len;
  }

  if(len > 4) exit(1);
  return len;
}


// Write the integer n using numBytes and starting from the Most 
// Significant Byte, and in the subarray t[*from,...].
// The parameter *from is changed accordingly: *from += numBytes.
void writeint(int n, int numBytes, unsigned char *t, int *from)
{	

  // Sanity check !!
  if (numBytes > 4){
    fprintf(stderr, "You requested to write more than 5 bytes of an Int\n\n");
    exit(1);  }

  if(numBytes >=4) {
    t[*from] = (n >> 24) & 0xFF;
    *from += 1; }

  if(numBytes >=3) {
    t[*from] = (n >> 16) & 0xFF;
    *from += 1; }

  if(numBytes >=2) {
    t[*from] = (n >> 8) & 0xFF;
    *from += 1; }

  t[*from] = n & 0xFF;
  *from += 1;

}

// Modulus: a mod b
long mod(double a, long b)
{	long c = (int) (a / (double) b);
	return (long) (a - ((double) b * (double) c));
}



/* ---------------- MAIN TYPES AND DATA ---------------------- */

typedef unsigned long SigType;	 /* Rabin-Karp fingerprint signature */

#define CHARSIZE 256

static unsigned char *text;		       /* Text to compress: array */
static int textSize;			       /* Text to compress: size */
static int blockSize;			       /* substring to hash: (minimum) size */
static int clear=0;		               /* Start of uncompressed text */


/* ------------------- SEARCH FOR GOOD MATCHES ---------------------- */
static int bestlen;		/* Longest match found: length */
static int bestoldstart;	/* Start in transmitted portion */
static int bestnewstart;	/* Start after text[clear] */

void expandmatch(int m1, int m2) /* m1 = new, m2 = old */
{	
  int i, len;


  m1 = m1 - blockSize +1;	/* Point at start of strings, */
  m2 = m2 - blockSize +1;	/*    not ends */

  if((m1 < 0) || (m2<0))
    return;

  // Checks if it is not a false match
  for (i = 0; i < blockSize; i++)
    if (text[m1+i] != text[m2+i])
      return;

  // Expands to the left, avoiding overlap with previous match
  len = blockSize;
  while ((m1 > clear) && (text[m1-1] == text[m2-1])) {
    m1--; m2--; len++;
  }

  // Expands to the right
  while (text[m1+len] == text[m2+len])
    len++;

  // Updates the best match
  // Since we update the hashTable by appending to the head
  //   in case of equal-length matches we take the closest one
  if (len > bestlen) {
    bestlen = len;
    bestnewstart = m1;
    bestoldstart = m2;
  }
}

// Copies text[n1,n2] into t[*from,....] possibly
// re-encoding the symbol <, and updates *from
void writeseq(int n1, int n2, unsigned char *t, int *from)
{	int i;

	for (i = n1; i <= n2; i++) {
	  if (text[i] == '<'){
	    t[*from] = text[i];	/* << is quoted < */
	    *from += 1;
	    t[*from] = 0;	/* special value */
	    *from += 1;
	  } else{
	    t[*from] = text[i];	/* << is quoted < */
	    *from += 1;
	  }
	}
}



/* --------------------------- HASH TABLE ------------------------------ */

typedef struct hnode *Hptr;
typedef struct hnode {
	Hptr	next;
	SigType	sig;
	int	pos;
} Hnode;

#define HSIZE ((1<<24) + 1)         /* Size of the hash table 2^24+1 */
static Hptr htab[HSIZE];                   /* Pointer to the Hash Table */

void hashInit()
{	
  int i;
  for(i=0; i < HSIZE; i++)
    htab[i] = (Hptr) 0;

}

void hashFree()
{	
  int i;
  Hptr p,q;

  for (i = 0; i < HSIZE; i++)
    if (htab[i]){
      p = htab[i];
      while(p) { q = p; p = p->next;; free(q); }
      htab[i] = (Hptr) 0;
    }
}


// ----------------------------------------------
// We left the BmI's choice of having dynamic allocation
// of hash-lists, a newer implementation should consider
// a static allocation, probably using two hash functions:
// one weaker but faster to be computed, and the other 
// stronger but slower (a la' rsync).
void store(SigType s, int i)
{	
  int j = s % HSIZE;
  Hptr p = (Hptr) malloc(sizeof(Hnode));
  if (p == 0) {
    printf("Malloc died in HashTable allocation\n\n");
    exit(1);
  }
  p->next = htab[j];  // Append to the head
  htab[j] = p;
  p->sig = s;
  p->pos = i;
}

// Check the current chunk-signature for a match
void checksig(SigType s, int i, unsigned char *t, int *from)
{	
  Hptr p;

  bestlen = -1;
  for (p = htab[s%HSIZE]; p; p = p->next)
    if (p->sig == s)
      expandmatch(i, p->pos);

  if (bestlen >= blockSize) {

    // writes the uncopied part text[clear, bestnewstart-1]
    // and changes *from
    writeseq(clear, bestnewstart-1, t, from); 

    // Writes the copied part: unlike BMI, we store the difference, as usual
    // dummy sequence is <X then follows the encoding of the dist+len in 
    // X/4 bytes for dist and X%4 for len.
    t[*from] = '<';
    *from += 1;

    // This encodes the len of the encoding of dist&len
    t[*from] = GetByteLen(bestnewstart - bestoldstart) * 4 + GetByteLen(bestlen);
    *from += 1;

    // sanity check !!
    if(t[*from-1] > 16){
      fprintf(stderr,"Encoding the dist+len: too large %d\n",t[*from]);
      exit(1); }



    // just the difference
    int tmpLen = GetByteLen(bestnewstart - bestoldstart);
    writeint((int) bestnewstart - bestoldstart, tmpLen, t, from);
    tmpLen = GetByteLen(bestlen);
    writeint((int) bestlen, tmpLen, t, from);

    clear = bestnewstart + bestlen;
  }
}




/* ---------------------  COMPRESSOR ----------------------- */
// We have left BmI's use of global variables, turning them to
// "static" in order to avoid conflicts with the other software 
// modules. The sytax is zlib-like one, with the addition of "b"
// to take care of the chuck-size to be hashed.
void bmi_compress(unsigned char *dest, int *destLen, unsigned char *src, int srcLen, int b)
{


  // Bentley-McIlroy's variables
  int i;
  int j, c, oldc;
  SigType s = 0, db; 
  SigType d = 32, q = 1999999973;
  int bmiCurr;


  clear = 0;
  text = src;
  textSize = srcLen;
  blockSize = b;

  
  /* Karp-Rabin's signatur needs: d^(b - 1) mod q */
  db = 1;
  for (j = 1; j < blockSize; j++) {
    /* db = (db * d) % q; */
    db = mod((double) db * d, q);  }
  

  hashInit();
  bmiCurr = 0;      // iterator over the bmiArray being written
  for (i = 0; i < srcLen; i++) {
    oldc = (i-blockSize >= 0 ? src[i - blockSize] : 0); // manage prec src
    c = src[i];
    s = mod(s + CHARSIZE * (double) q - oldc * (double) db, q);
    s = mod(s * (double) d + c, q);
    
    // starts checking only if the signature 
    // refers to an uncompressed block
    if ((i - blockSize + 1) >= clear)
      checksig(s, i, dest, &bmiCurr);
    
    // Stores only signatures at %b offsets
    if (i % blockSize == blockSize-1)
      store(s, i);
  }
  
  // Writes the last part which has been not copied
  writeseq(clear, srcLen-1, dest, &bmiCurr);

  if (bmiCurr > *destLen){
    printf("Error in writing BMI-output (%d)\n",bmiCurr);
    exit(1);
  } else { *destLen = bmiCurr;  }

  hashFree();
}
  
